@@ -41,7 +41,6 @@ from copy import copy
 from typing import Iterable, List, Optional, Sequence, Tuple, TypeVar
 
 import sympy as sp
-import numpy as np
 
 from pystencils.cache import memorycache
 from pystencils.sympyextensions import remove_higher_order_terms
@@ -160,7 +159,7 @@ def polynomial_to_exponent_representation(polynomial, dim=3):
     summands = [polynomial] if polynomial.func != sp.Add else polynomial.args
     for expr in summands:
         if len(expr.atoms(sp.Symbol) - set(MOMENT_SYMBOLS)) > 0:
-            raise ValueError("Invalid moment polynomial: " + str(expr))
+            raise ValueError(f"Invalid moment polynomial: {str(expr)}")
         c, x_exp, y_exp, z_exp = sp.Wild('c'), sp.Wild('xexp'), sp.Wild('yexp'), sp.Wild('zc')
         match_res = expr.match(c * x**x_exp * y**y_exp * z**z_exp)
         assert match_res[x_exp].is_integer and match_res[y_exp].is_integer and match_res[z_exp].is_integer
@@ -180,6 +179,10 @@ def moment_sort_key(moment):
     return get_order(moment), len(moment.atoms(sp.Symbol)), len(mom_str), mom_str
 
 
+def exponent_tuple_sort_key(x):
+    return moment_sort_key(exponent_to_polynomial_representation(x))
+
+
 def sort_moments_into_groups_of_same_order(moments):
     """Returns a dictionary mapping the order (int) to a list of moments with that order."""
     result = defaultdict(list)
@@ -191,6 +194,10 @@ def sort_moments_into_groups_of_same_order(moments):
 # -------------------- Common Function working with exponent tuples and polynomial moments -----------------------------
 
 
+def statistical_quantity_symbol(name, exponents):
+    return sp.Symbol(f'{name}_{"".join(str(i) for i in exponents)}')
+
+
 def is_even(moment):
     """
     A moment is considered even when under sign reversal nothing changes i.e. :math:`m(-x,-y,-z) = m(x,y,z)`
@@ -282,6 +289,62 @@ def non_aliased_moment(moment_tuple: Sequence[int]) -> Tuple[int, ...]:
     return tuple(result)
 
 
+def aliases_from_moment_list(moment_exponents, stencil):
+    """Takes a list of moment exponent tuples and finds aliases within it
+    according the given stencil. Two moments are aliases of each other on a stencil
+    if they produce the same coefficients in the moment sum over all populations.
+
+    Apart from the obvious aliases (e.g. ``(4,0,0)`` to ``(2,0,0)``, etc), there are aliasing
+    effects for example on the D3Q15 stencil, where some third and fourth order moments
+    are the same.
+    """
+    mm = moment_matrix(moment_exponents, stencil)
+    rows_dict = dict()
+    aliases = dict()
+    for r, moment in enumerate(moment_exponents):
+        row = tuple(mm[r, :])
+        if row in rows_dict:
+            aliases[moment] = rows_dict[row]
+        else:
+            rows_dict[row] = moment
+    return aliases
+
+
+def non_aliased_polynomial_raw_moments(polys, stencil, nested=False):
+    """Takes a (potentially nested) list of raw moment polynomials and rewrites them by eliminating
+    any aliased monomials. 
+
+    All polynomials are expanded and occuring monomials are collected. Using `aliases_from_moment_list`,
+    aliases are eliminated and substituted in the polynomials. 
+
+    Attention: Only use this method for monomials in raw moment space. It will produce wrong results
+    if used for central moments, since there is no direct aliasing in central moment space!
+    """
+    dim = len(stencil[0])
+    if nested:
+        polys_unnested = list(itertools.chain.from_iterable(polys))
+    else:
+        polys_unnested = polys
+    monos = sorted(extract_monomials(polys_unnested, dim=dim), key=exponent_tuple_sort_key)
+    aliases = aliases_from_moment_list(monos, stencil)
+
+    if not aliases:  # Stop early if there are no aliases
+        return polys
+
+    def nonalias_polynomial(poly):
+        exponents = polynomial_to_exponent_representation(poly, dim)
+        exponents_unaliased = [(coeff, aliases.get(m, m)) for coeff, m in exponents]
+        return sum(coeff * exponent_to_polynomial_representation(m) for coeff, m in exponents_unaliased)
+
+    if nested:
+        output_polys = []
+        for group in polys:
+            output_polys.append(list(map(nonalias_polynomial, group)))
+        return output_polys
+    else:
+        return list(map(nonalias_polynomial, polys))
+
+
 def is_bulk_moment(moment, dim):
     """The bulk moment is x**2+y**2+z**2"""
     if type(moment) is not tuple:
@@ -330,10 +393,10 @@ def discrete_moment(func, moment, stencil, shift_velocity=None):
         func: list of distribution functions for each direction
         moment: can either be a exponent tuple, or a sympy polynomial expression
         stencil: sequence of directions
-        shift_velocity: velocity vector u to compute central moments, the lattice velocity is replaced by
-                        (lattice_velocity - shift_velocity)
+        shift_velocity: velocity vector :math:`\mathbf{u}` to compute central moments, the lattice
+                        velocity is replaced by (lattice_velocity - shift_velocity)
     """
-    assert len(stencil) == len(func)
+    assert stencil.Q == len(func)
     if shift_velocity is None:
         shift_velocity = (0,) * len(stencil[0])
     res = 0
@@ -374,10 +437,10 @@ def moment_matrix(moments, stencil, shift_velocity=None):
                 evaluated = evaluated.subs(var, stencil_entry - shift)
             return evaluated
 
-    return sp.Matrix(len(moments), len(stencil), generator)
+    return sp.Matrix(len(moments), stencil.Q, generator)
 
 
-def set_up_shift_matrix(moments, stencil, velocity_symbols=None):
+def set_up_shift_matrix(moments, stencil, velocity_symbols=sp.symbols("u_:3")):
     """
     Sets up a shift matrix to shift raw moments to central moment space.
 
@@ -387,50 +450,20 @@ def set_up_shift_matrix(moments, stencil, velocity_symbols=None):
         - stencil: Nested tuple of lattice velocities
         - velocity_symbols: Sequence of symbols corresponding to the shift velocity
     """
-    x, y, z = MOMENT_SYMBOLS
     dim = len(stencil[0])
-    nr_directions = len(stencil)
+    if len(velocity_symbols) > dim:
+        velocity_symbols = velocity_symbols[:dim]
 
-    directions = np.asarray(stencil)
+    M = moment_matrix(moments, stencil, shift_velocity=None)
+    MN = moment_matrix(moments, stencil, shift_velocity=velocity_symbols)
 
-    u = velocity_symbols if velocity_symbols is not None else sp.symbols(f"u_:{dim}")
-    f = sp.symbols(f"f_:{nr_directions}")
-    m = sp.symbols(f"m_:{nr_directions}")
+    N = sp.simplify(MN * M.inv())
 
-    Mf = moment_matrix(moments, stencil=stencil) * sp.Matrix(f)
+    assert N.is_lower, "Calculating the shift matrix gave not a lower diagonal matrix. Thus it failed"
+    assert sum(N[i, i] for i in range(stencil.Q)) == stencil.Q, "Calculating the shift matrix failed. " \
+        "There are entries on the diagonal which are not equal to one"
 
-    shift_matrix_list = []
-    for nr_moment in range(len(moments)):
-        if not isinstance(moments[nr_moment], tuple):
-            exponent_central_moment = moments[nr_moment].as_powers_dict()
-            exponent = [exponent_central_moment[x], exponent_central_moment[y], exponent_central_moment[z]]
-        else:
-            exponent = moments[nr_moment]
-
-        shift_equation = 1
-        for i in range(dim):
-            shift_equation *= (directions[:, i] - u[i]) ** exponent[i]
-        shift_equation = sum(shift_equation * f)
-
-        collected_expr = sp.Poly(shift_equation, u).as_expr()
-        terms_of_collected_expr = collected_expr.args
-
-        constants = set()
-        for i in range(len(terms_of_collected_expr)):
-            constants.add(sp.Abs(sp.LC(terms_of_collected_expr[i])))
-
-        for i in range(len(stencil)):
-            for c in constants:
-                collected_expr = collected_expr.subs(c * Mf[i], c * m[i])
-
-        collected_expr = sp.collect(collected_expr.expand(), m)
-
-        inner_list = []
-        for i in range(len(stencil)):
-            inner_list.append(collected_expr.coeff(m[i], 1))
-
-        shift_matrix_list.append(inner_list)
-    return sp.Matrix(shift_matrix_list)
+    return N
 
 
 def gram_schmidt(moments, stencil, weights=None):
@@ -441,15 +474,15 @@ def gram_schmidt(moments, stencil, weights=None):
         moments: sequence of moments, either in tuple or polynomial form
         stencil: stencil as sequence of directions
         weights: optional weights, that define the scalar product which is used for normalization.
-                 Scalar product :math:`< a,b > = \sum a_i b_i w_i` with weights :math:`w_i`.
+                 Scalar product :math:`< \mathbf{a},\mathbf{b} > = \sum a_i b_i w_i` with weights :math:`w_i`.
                  Passing no weights sets all weights to 1.
     Returns:
         set of orthogonal moments in polynomial form
     """
     if weights is None:
-        weights = sp.eye(len(stencil))
+        weights = sp.eye(stencil.Q)
     if type(weights) is list:
-        assert len(weights) == len(stencil)
+        assert len(weights) == stencil.Q
         weights = sp.diag(*weights)
 
     if type(moments[0]) is tuple:
@@ -466,8 +499,8 @@ def gram_schmidt(moments, stencil, weights=None):
             prev_element = orthogonalized_vectors[j]
             denominator = prev_element.dot(weights * prev_element)
             if denominator == 0:
-                raise ValueError("Not an independent set of vectors given: "
-                                 "vector %d is dependent on previous vectors" % (i,))
+                raise ValueError(f"Not an independent set of vectors given: "
+                                 f"vector {i} is dependent on previous vectors")
             overlap = current_element.dot(weights * prev_element) / denominator
             current_element -= overlap * prev_element
             moments[i] -= overlap * moments[j]
@@ -480,22 +513,19 @@ def get_default_moment_set_for_stencil(stencil):
     """
     Returns a sequence of moments that are commonly used to construct a LBM equilibrium for the given stencil
     """
-    from lbmpy.stencils import get_stencil
-    from pystencils.stencil import have_same_entries
-
     to_poly = exponents_to_polynomial_representations
 
-    if have_same_entries(stencil, get_stencil("D2Q9")):
+    if stencil.D == 2 and stencil.Q == 9:
         return sorted(to_poly(moments_up_to_component_order(2, dim=2)), key=moment_sort_key)
 
     all27_moments = moments_up_to_component_order(2, dim=3)
-    if have_same_entries(stencil, get_stencil("D3Q27")):
+    if stencil.D == 3 and stencil.Q == 27:
         return sorted(to_poly(all27_moments), key=moment_sort_key)
-    if have_same_entries(stencil, get_stencil("D3Q19")):
+    if stencil.D == 3 and stencil.Q == 19:
         non_matched_moments = [(1, 2, 2), (1, 1, 2), (2, 2, 2), (1, 1, 1)]
         moments19 = set(all27_moments) - set(extend_moments_with_permutations(non_matched_moments))
         return sorted(to_poly(moments19), key=moment_sort_key)
-    if have_same_entries(stencil, get_stencil("D3Q15")):
+    if stencil.D == 3 and stencil.Q == 15:
         x, y, z = MOMENT_SYMBOLS
         non_matched_moments = [(1, 2, 0), (2, 2, 0), (1, 1, 2), (1, 2, 2), (2, 2, 2)]
         additional_moments = (6 * (x ** 2 * y ** 2 + x ** 2 * z ** 2 + y ** 2 * z ** 2),
@@ -505,6 +535,15 @@ def get_default_moment_set_for_stencil(stencil):
                               )
         to_remove = set(extend_moments_with_permutations(non_matched_moments))
         return sorted(to_poly(set(all27_moments) - to_remove) + additional_moments, key=moment_sort_key)
+    if stencil.D == 3 and stencil.Q == 7:
+        x, y, z = MOMENT_SYMBOLS
+        non_matched_moments = [(1, 1, 0), (1, 1, 1), (1, 1, 2), (1, 2, 0),
+                               (1, 2, 2), (2, 0, 0), (2, 2, 0), (2, 2, 2)]
+        additional_moments = (x ** 2 - y ** 2,
+                              x ** 2 - z ** 2,
+                              x ** 2 + y ** 2 + z ** 2)
+        to_remove = set(extend_moments_with_permutations(non_matched_moments))
+        return sorted(to_poly(set(all27_moments) - to_remove) + additional_moments, key=moment_sort_key)
 
     raise NotImplementedError("No default moment system available for this stencil - define matched moments yourself")
 
@@ -539,8 +578,7 @@ def monomial_to_polynomial_transformation_matrix(monomials, polynomials):
         polynomials: sequence of polynomials in the MOMENT_SYMBOLS
 
     >>> x, y, z = MOMENT_SYMBOLS
-    >>> polys = [7 * x**2 + 3 * x + 2 * y **2, \
-                 9 * x**2 - 5 * x]
+    >>> polys = [7 * x**2 + 3 * x + 2 * y **2, 9 * x**2 - 5 * x]
     >>> mons = list(extract_monomials(polys, dim=2))
     >>> mons.sort()
     >>> monomial_to_polynomial_transformation_matrix(mons, polys)
@@ -558,6 +596,33 @@ def monomial_to_polynomial_transformation_matrix(monomials, polynomials):
     return result
 
 
+def central_moment_reduced_monomial_to_polynomial_matrix(polynomials, stencil, velocity_symbols=None):
+    r"""
+    Returns a transformation matrix from a reduced set of monomial central moments to a set of polynomial
+    central moments.
+
+    Use for a set of :math:`q` central moment polynomials that reduces to a too-large set of :math:`r > q` 
+    monomials (as given by `extract_monomials`). Reduces the monomials by eliminating aliases in raw moment
+    space, and computes a reduced polynomialization matrix :math:`\mathbf{P}^{r}` that computes the given
+    polynomials from the reduced set of monomials.
+    """
+    dim = len(stencil[0])
+    if velocity_symbols is None:
+        velocity_symbols = sp.symbols(f"u_:{dim}")
+
+    reduced_polynomials = non_aliased_polynomial_raw_moments(polynomials, stencil)
+    reduced_monomials = sorted(extract_monomials(reduced_polynomials), key=exponent_tuple_sort_key)
+
+    assert len(reduced_monomials) == stencil.Q, "Could not extract a base set of correct size from the polynomials."
+
+    N = set_up_shift_matrix(reduced_monomials, stencil, velocity_symbols=velocity_symbols)
+    N_P = set_up_shift_matrix(polynomials, stencil, velocity_symbols=velocity_symbols)
+    P_mono = monomial_to_polynomial_transformation_matrix(reduced_monomials, reduced_polynomials)
+
+    P_reduced = (N_P * P_mono * N.inv()).expand()
+    return P_reduced, reduced_monomials
+
+
 # ---------------------------------- Visualization ---------------------------------------------------------------------
 
 
@@ -581,21 +646,20 @@ def moment_equality_table(stencil, discrete_equilibrium=None, continuous_equilib
     import ipy_table
     from lbmpy.continuous_distribution_measures import continuous_moment
 
-    dim = len(stencil[0])
-    u = sp.symbols("u_:{dim}".format(dim=dim))
+    u = sp.symbols(f"u_:{stencil.D}")
     if discrete_equilibrium is None:
         from lbmpy.maxwellian_equilibrium import discrete_maxwellian_equilibrium
         discrete_equilibrium = discrete_maxwellian_equilibrium(stencil, c_s_sq=sp.Rational(1, 3), compressible=True,
                                                                u=u, order=truncate_order)
     if continuous_equilibrium is None:
         from lbmpy.maxwellian_equilibrium import continuous_maxwellian_equilibrium
-        continuous_equilibrium = continuous_maxwellian_equilibrium(dim=dim, u=u, c_s_sq=sp.Rational(1, 3))
+        continuous_equilibrium = continuous_maxwellian_equilibrium(dim=stencil.D, u=u, c_s_sq=sp.Rational(1, 3))
 
     table = []
     matched_moments = 0
     non_matched_moments = 0
 
-    moments_list = [list(moments_of_order(o, dim, include_permutations=False)) for o in range(max_order + 1)]
+    moments_list = [list(moments_of_order(o, stencil.D, include_permutations=False)) for o in range(max_order + 1)]
 
     colors = dict()
     nr_of_columns = max([len(v) for v in moments_list]) + 1
@@ -606,11 +670,11 @@ def moment_equality_table(stencil, discrete_equilibrium=None, continuous_equilib
 
     for order, moments in enumerate(moments_list):
         row = [' '] * nr_of_columns
-        row[0] = '%d' % (order,)
+        row[0] = f'{order}'
         for moment, col_idx in zip(moments, range(1, len(row))):
             multiplicity = moment_multiplicity(moment)
             dm = discrete_moment(discrete_equilibrium, moment, stencil)
-            cm = continuous_moment(continuous_equilibrium, moment, symbols=sp.symbols("v_0 v_1 v_2")[:dim])
+            cm = continuous_moment(continuous_equilibrium, moment, symbols=sp.symbols("v_0 v_1 v_2")[:stencil.D])
             difference = sp.simplify(dm - cm)
             if truncate_order:
                 difference = sp.simplify(remove_higher_order_terms(difference, symbols=u, order=truncate_order))
@@ -621,7 +685,7 @@ def moment_equality_table(stencil, discrete_equilibrium=None, continuous_equilib
                 colors[(order + 1, col_idx)] = 'lightGreen'
                 matched_moments += multiplicity
 
-            row[col_idx] = '%s  x %d' % (moment, moment_multiplicity(moment))
+            row[col_idx] = f'{moment}  x {moment_multiplicity(moment)}'
 
         table.append(row)
 
@@ -630,8 +694,8 @@ def moment_equality_table(stencil, discrete_equilibrium=None, continuous_equilib
     for cell_idx, color in colors.items():
         ipy_table.set_cell_style(cell_idx[0], cell_idx[1], color=color)
 
-    print("Matched moments %d - non matched moments %d - total %d" %
-          (matched_moments, non_matched_moments, matched_moments + non_matched_moments))
+    print(f"Matched moments {matched_moments} - non matched moments {non_matched_moments} "
+          f"- total {matched_moments + non_matched_moments}")
 
     return table_display
 
@@ -662,8 +726,8 @@ def moment_equality_table_by_stencil(name_to_stencil_dict, moments, truncate_ord
 
     colors = {}
     for stencil_idx, stencil in enumerate(stencils):
-        dim = len(stencil[0])
-        u = sp.symbols("u_:{dim}".format(dim=dim))
+        dim = stencil.D
+        u = sp.symbols(f"u_:{dim}")
         discrete_equilibrium = discrete_maxwellian_equilibrium(stencil, c_s_sq=sp.Rational(1, 3), compressible=True,
                                                                u=u, order=truncate_order)
         continuous_equilibrium = continuous_maxwellian_equilibrium(dim=dim, u=u, c_s_sq=sp.Rational(1, 3))
@@ -720,7 +784,7 @@ def __unique_permutations(elements: Sequence[T]) -> Iterable[T]:
 
     """
     if len(elements) == 1:
-        yield (elements[0],)
+        yield elements[0],
     else:
         unique_elements = set(elements)
         for first_element in unique_elements:
